/* COMPRESSOR - Audio WAV file dynamic compressor  *
 * RELEASE 0.30 - V.PORTET - 2005-2018   -  From Windows x64 Release Package */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <time.h>
#include <ctype.h>

#define fdbthreshold 35
#define lowfreql ((float)0.013)

typedef union
	{
	signed short calcsamp;
	struct
		{
		signed char lsbyte;
		signed char msbyte;
		} sampclip;
	} sampful;
sampful lowchan_l;
sampful lowchan_l_primary;
sampful highchan_l;
sampful inchan_l;
sampful lowchan_r;
sampful lowchan_r_primary;
sampful highchan_r;
sampful inchan_r;
//
int dumm;
int i;
char filname[126];
char tmpin[3];
signed char *source;
unsigned long sourcel;
unsigned long sourcepointer;
float maxgain;
float recoverspeed;
signed char temp;
unsigned int fdbspeed;
float fastpump = 1;
unsigned char pumprec = 0;
//
unsigned char qpdet;
unsigned char cntlp;
float lgain = 0.4;
float rgain = 0.4;
float targain = 0.4;
signed char lamp;
signed char ramp;
//
unsigned char qpdet_lfc;
unsigned char cntlp_lfc;
float lgain_lfc = 0.6;
float rgain_lfc = 0.6;
float targain_lfc = 0.6;
signed char lamp_lfc;
signed char ramp_lfc;
//


FILE *wavopen;

int main(int argc,char *argv[])
	{
	filname[0]=0;
	printf("\nAudio Compressor V0.30 - V.Portet 2005-2018\n\n");
	if (argc==1)
		{
		while (strlen(filname)<5)
			{
			printf("\nWAV File to be processed : ");
			scanf("%124s",filname);
			}
		printf("\n\n");
		maxgain = 20;
		recoverspeed = 8;
		fdbspeed = 2;
		printf("Using default settings:\n");
		}
	else
		{
		strncpy(filname,argv[1],124);
		if (argc!=5)
			{
			maxgain = 20;
			recoverspeed = 8;
			fdbspeed = 2;
			printf("Using default settings:\n");
			}
		else
			{
			sscanf(argv[2],"%f",&maxgain);
			sscanf(argv[3],"%f",&recoverspeed);
			sscanf(argv[4],"%d",&fdbspeed);
			if ((maxgain>=4)&&(maxgain<=100)&&(recoverspeed>=1)&&(recoverspeed<=9)&&(fdbspeed>=1)&&(fdbspeed<=4))
				{
				printf("User defined settings:\n");
				}
			else
				{
				maxgain = 20;
				recoverspeed = 8;
				fdbspeed = 2;
				printf("Using default settings:\n");
				}
			}
		}
	printf("Max Gain (4-100) = %f\n",maxgain);
	printf("Recover Speed (1-9) = %f\n",recoverspeed);
	printf("Feedback Speed (1-4) = %d\n",fdbspeed);
	printf("\n\n");
	wavopen=fopen(filname,"rb");
	if (wavopen)
		{
		fseek(wavopen,0,SEEK_END);
		sourcel = ftell(wavopen);
		if (sourcel < 20000)
			{
			printf("This sample is too short for being compressed !\n");
			fclose(wavopen);
			}
		else
			{
			source = malloc(sourcel);
			if (source==NULL)
				{
				printf("Not enough available memory !\n");
				fclose(wavopen);
				}
			else
				{
				printf("Processing...\n\n");
				// Loading data
				fseek(wavopen,0,SEEK_SET);
				fread(&source[0],sourcel,1,wavopen);
				// 1st - Preset start values for gains
				lamp = 0;
				ramp = 0;
				lamp_lfc = 0;
				ramp_lfc = 0;
				lowchan_l.calcsamp = 0;
				lowchan_l_primary.calcsamp = 0;
				highchan_l.calcsamp = 0;
				lowchan_r.calcsamp = 0;
				lowchan_r_primary.calcsamp = 0;
				highchan_r.calcsamp = 0;
				qpdet = 0;
				cntlp = 0;
				qpdet_lfc = 0;
				cntlp_lfc = 0;
				for (sourcepointer = 44; sourcepointer < sourcel; sourcepointer += 4)
					{
					// 2-band compression => Filter first
					// Left
					inchan_l.sampclip.lsbyte = source[sourcepointer];
					inchan_l.sampclip.msbyte = source[sourcepointer+1];
					lowchan_l_primary.calcsamp += (signed int)((float)(inchan_l.calcsamp - lowchan_l_primary.calcsamp)*lowfreql);
					highchan_l.calcsamp = inchan_l.calcsamp - lowchan_l_primary.calcsamp;
					// Right
					inchan_r.sampclip.lsbyte = source[sourcepointer+2];
					inchan_r.sampclip.msbyte = source[sourcepointer+3];
					lowchan_r_primary.calcsamp += (signed int)((float)(inchan_r.calcsamp - lowchan_r_primary.calcsamp)*lowfreql);
					highchan_r.calcsamp = inchan_r.calcsamp - lowchan_r_primary.calcsamp;
					// Most significant bytes can then be read directly from filtered channel
					// Apply current gain values
					highchan_l.calcsamp = (signed int)(lgain * fastpump * (float)highchan_l.calcsamp);
					highchan_r.calcsamp = (signed int)(rgain * fastpump * (float)highchan_r.calcsamp);
					lowchan_l.calcsamp = (signed int)(lgain_lfc * (float)lowchan_l_primary.calcsamp);
					lowchan_r.calcsamp = (signed int)(rgain_lfc * (float)lowchan_r_primary.calcsamp);
					// Amplitude peak demodulation - high F channel
					temp = abs(highchan_l.sampclip.msbyte);
					if (temp>lamp)
						{
						lamp=temp;
						qpdet = 0;
						}
					temp = abs(highchan_r.sampclip.msbyte);
					if (temp>ramp)
						{
						ramp=temp;
						qpdet = 0;
						}
					qpdet += 1;
					if (qpdet >= 50)
						{
						if (lamp>0)
							{
							lamp -= 1;
							}
						if (ramp>0)
							{
							ramp -= 1;
							}
						qpdet = 0;
						}
					// Amplitude peak demodulation - Low F channel
					temp = abs(lowchan_l.sampclip.msbyte);
					if (temp>lamp_lfc)
						{
						lamp_lfc=temp;
						qpdet_lfc = 0;
						}
					temp = abs(lowchan_r.sampclip.msbyte);
					if (temp>ramp_lfc)
						{
						ramp_lfc=temp;
						qpdet_lfc = 0;
						}
					qpdet_lfc += 1;
					if (qpdet_lfc >= 50)
						{
						if (lamp_lfc>0)
							{
							lamp_lfc -= 1;
							}
						if (ramp_lfc>0)
							{
							ramp_lfc -= 1;
							}
						qpdet_lfc = 0;
						}
					// Gain control - High F channel
					if (lamp != 0)
						{
						targain = (float)fdbthreshold * lgain / ((float) lamp);
						if (targain < (float)0.2)
							{
							targain = 0.2;
							}
						}
					else
						{
						targain = maxgain;
						}
					if (targain<lgain)
						{
						if (targain<(0.7*lgain))
							{
							lgain -= (0.025 * (lgain-targain) * (float)fdbspeed);
							lamp = fdbthreshold;
							}
						else
							{
							lgain -= (0.001 * (lgain-targain) * (float)fdbspeed);
							}
						}
					else
						{
						if (qpdet==0)
							{
							if (lgain < maxgain)
								{
								lgain *= (1 + (recoverspeed/(float)500));
								}
							}
						}
					if (ramp != 0)
						{
						targain = (float)fdbthreshold * rgain / ((float) ramp);
						if (targain < (float)0.2)
							{
								targain = 0.2;
							}
						}
					else
						{
						targain = maxgain;
						}
					if (targain<rgain)
						{
						if (targain<(0.7*rgain))
							{
							rgain -= (0.025 * (rgain-targain) * (float)fdbspeed);
							ramp = fdbthreshold;
							}
						else
							{
							rgain -= (0.001 * (rgain-targain) * (float)fdbspeed);
							}
						}
					else
						{
						if (qpdet==0)
							{
							if (rgain < maxgain)
								{
								rgain *= (1 + (recoverspeed/(float)500));
								}
							}
						}
					// Gain control - Low F channel
					if (lamp_lfc != 0)
						{
						targain_lfc = (float)(1.5*fdbthreshold) * lgain_lfc / ((float) lamp_lfc);
						if (targain_lfc < (float)0.2)
							{
							targain_lfc = 0.2;
							}
						}
					else
						{
						targain_lfc = maxgain;
						}
					if (targain_lfc<lgain_lfc)
						{
						if (targain_lfc<(0.7*lgain_lfc))
							{
							lgain_lfc -= (0.05 * (lgain_lfc - targain_lfc));
							lamp_lfc = fdbthreshold;
							}
						else
							{
							lgain_lfc -= (0.001 * (lgain_lfc - targain_lfc));
							}
						}
					else
						{
						if (qpdet_lfc==0)
							{
							if (lgain_lfc < maxgain)
								{
								lgain_lfc *= (1 + (recoverspeed/800));
								}
							}
						}
					if (ramp_lfc != 0)
						{
						targain_lfc = (float)(1.5*fdbthreshold) * rgain_lfc / ((float) ramp_lfc);
						if (targain_lfc < (float)0.2)
							{
							targain_lfc = 0.2;
							}
						}
					else
						{
						targain_lfc = maxgain;
						}
					if (targain_lfc<rgain_lfc)
						{
						if (targain_lfc<(0.7*rgain_lfc))
							{
							rgain_lfc -= (0.05 * (rgain_lfc - targain_lfc));
							ramp_lfc = fdbthreshold;
							}
						else
							{
							rgain_lfc -= (0.001 * (rgain_lfc - targain_lfc));
							}
						}
					else
						{
						if (qpdet_lfc==0)
							{
							if (rgain_lfc < maxgain)
								{
								rgain_lfc *= (1 + (recoverspeed/800));
								}
							}
						}
					// Recover from possible "fast pumping" compression (i.e. bass priority)
					pumprec++;
					if (((pumprec&0x07)==0)&&(fastpump < (float)1))
						{
						fastpump *= 1.05;
						}
					// Mix bands + Update in buffer with clipping and "fast pumping" protections
					// (The "pumping" factor is estimated from the fact that thresholds for low and high bands are respectively 2/3 and 1/3 when not clipping)
					inchan_l.calcsamp = (highchan_l.calcsamp / 16) + (lowchan_l.calcsamp / 16);
					if (inchan_l.calcsamp > (signed int)2047)
						{
						fastpump = (float)1 - ((float)0.00146 * (float)(inchan_l.calcsamp - (signed int)2047));
						if (fastpump < 0.1)
							{
							fastpump = 0.1;
							}
						inchan_l.calcsamp = 32767;
						}
					else
						{
						if (inchan_l.calcsamp < (signed int)(-2048))
							{
							fastpump = (float)1 + ((float)0.00146 * (float)(inchan_l.calcsamp + (signed int)2048));
							if (fastpump < 0.1)
								{
								fastpump = 0.1;
								}
							inchan_l.calcsamp = -32768;
							}
						else
							{
							inchan_l.calcsamp = highchan_l.calcsamp + lowchan_l.calcsamp;
							}
						}						
					inchan_r.calcsamp = (highchan_r.calcsamp / 16) + (lowchan_r.calcsamp / 16);
					if (inchan_r.calcsamp > (signed int)2047)
						{
						fastpump = (float)1 - ((float)0.00146 * (float)(inchan_r.calcsamp - (signed int)2047));
						if (fastpump < 0.1)
							{
							fastpump = 0.1;
							}
						inchan_r.calcsamp = 32767;
						}
					else
						{
						if (inchan_r.calcsamp < (signed int)(-2048))
							{
							fastpump = (float)1 + ((float)0.00146 * (float)(inchan_r.calcsamp + (signed int)2048));
							if (fastpump < 0.1)
								{
								fastpump = 0.1;
								}
							inchan_r.calcsamp = -32768;
							}
						else
							{
							inchan_r.calcsamp = highchan_r.calcsamp + lowchan_r.calcsamp;
							}
						}						
					source[sourcepointer] = inchan_l.sampclip.lsbyte;
					source[sourcepointer + 1] = inchan_l.sampclip.msbyte;
					source[sourcepointer + 2] = inchan_r.sampclip.lsbyte;
					source[sourcepointer + 3] = inchan_r.sampclip.msbyte;
					// Progression info
					if ((sourcepointer & 0x0001FFFF)==0)
						{
						printf(".");
						cntlp += 1;
						if (cntlp==80)
							{
							cntlp=0;
							printf("\n");
							}
						}
					}
				printf("\n");
				printf("Updating file...\n");
				i=0;
				dumm=strlen(filname);
				while ((dumm!=0)&&(i==0))  // Parses file name to find the last "/" or "\"
					{
					dumm -= 1;
					if ((filname[dumm] == 47)||(filname[dumm] == 0x5C))
						{
						i=dumm; // Position found
						}
					}
				// On exit, i=0 means no "/" or "\", any other value is its position
				dumm=strlen(filname)+1;  // Prepare string shift of +1 (terminating NULL included)
				if (i == 0)
					{
					while (dumm!=0)
						{
						filname[dumm]=filname[dumm-1];
						dumm-=1;
						}
					filname[0]='C';
					}
				else
					{
					while (dumm!=i+1)
						{
						filname[dumm]=filname[dumm-1];
						dumm-=1;
						}
					filname[i+1]='C';
					}
				wavopen=fopen(filname,"wb");
				if (wavopen)
					{
					fwrite(&source[0],sourcel,1,wavopen);
					fclose(wavopen);
					printf("Done !\n");
					}
				else
					{
					printf("Could not overwrite %s !\n",filname);
					}
				}
			}
		}
	else
		{
		printf("Could not open %s !\n\n",filname);
		}
	return 0;
	}

